@@ -264,7 +264,6 @@ class FormalTypeDeclaration(FormalDeclaration):
 
 
 class PackageDeclaration(Declaration):
-    # pylint: disable=too-many-arguments
     def __init__(
         self,
         identifier: StrID,
@@ -326,7 +325,6 @@ class GenericPackageInstantiation(Declaration):
 
 
 class ObjectDeclaration(Declaration):
-    # pylint: disable=too-many-arguments
     def __init__(
         self,
         identifiers: Sequence[StrID],
@@ -400,8 +398,8 @@ class TypeDeclaration(Declaration):
 
 
 class ModularType(TypeDeclaration):
-    def __init__(self, identifier: StrID, modulus: Expr) -> None:
-        super().__init__(identifier)
+    def __init__(self, identifier: StrID, modulus: Expr, aspects: Sequence[Aspect] = None) -> None:
+        super().__init__(identifier, aspects=aspects or [])
         self.modulus = modulus
 
     @property
@@ -410,11 +408,12 @@ class ModularType(TypeDeclaration):
 
 
 class RangeType(TypeDeclaration):
-    def __init__(self, identifier: StrID, first: Expr, last: Expr, size: Expr) -> None:
-        super().__init__(identifier, aspects=[Size(size)])
+    def __init__(
+        self, identifier: StrID, first: Expr, last: Expr, aspects: Sequence[Aspect] = None
+    ) -> None:
+        super().__init__(identifier, aspects=aspects or [])
         self.first = first
         self.last = last
-        self.size = size
 
     @property
     def type_definition(self) -> str:
@@ -423,7 +422,7 @@ class RangeType(TypeDeclaration):
 
 class EnumerationType(TypeDeclaration):
     def __init__(
-        self, identifier: StrID, literals: Mapping[str, Optional[Number]], size: Expr = None,
+        self, identifier: StrID, literals: Mapping[ID, Optional[Number]], size: Expr = None,
     ) -> None:
         super().__init__(identifier, aspects=([Size(size)] if size else []))
         self.literals = (
@@ -435,7 +434,7 @@ class EnumerationType(TypeDeclaration):
 
     @property
     def type_definition(self) -> str:
-        literal_specification = ", ".join(self.literals.keys())
+        literal_specification = ", ".join(map(str, self.literals.keys()))
         return f" ({literal_specification})"
 
     @property
@@ -564,7 +563,6 @@ class VariantPart(Ada):
 
 
 class RecordType(TypeDeclaration):
-    # pylint: disable=too-many-arguments
     def __init__(
         self,
         identifier: StrID,
