@@ -10,6 +10,7 @@ import z3
 
 from rflx.common import generic_repr, indent, indent_next, unique
 from rflx.contract import DBC, invariant, require
+from rflx.error import Location
 from rflx.identifier import ID, StrID
 
 
@@ -48,7 +49,7 @@ class Proof:
         return self.__result
 
     @property
-    def error(self) -> str:
+    def error(self) -> List[Tuple[str, Optional[Location]]]:
         assert self.__result == ProofResult.unsat
         solver = z3.Solver()
         solver.set(unsat_core=True)
@@ -60,13 +61,21 @@ class Proof:
         facts["goal"] = self.__expr
         result = solver.check()
         assert result == z3.unsat, f"result should be unsat (is {result})"
-        return "\n   ∧ ".join([str(facts[str(fact)]) for fact in solver.unsat_core()])
+        return [
+            (" ".join(str(facts[str(fact)]).replace("\n", " ").split()), facts[fact].location)
+            for fact in sorted([str(h) for h in solver.unsat_core()])
+        ]
 
 
 class Expr(DBC):
+    def __init__(self, location: Location = None):
+        self.location = location
+
     def __eq__(self, other: object) -> bool:
         if isinstance(other, self.__class__):
-            return self.__dict__ == other.__dict__
+            return {k: v for k, v in self.__dict__.items() if k != "location"} == {
+                k: v for k, v in other.__dict__.items() if k != "location"
+            }
         return NotImplemented
 
     def __hash__(self) -> int:
@@ -107,8 +116,8 @@ class Expr(DBC):
     def precedence(self) -> Precedence:
         raise NotImplementedError
 
-    # pylint: disable=unused-argument,no-self-use
-    def variables(self, proof: bool = False) -> List["Variable"]:
+    # pylint: disable=no-self-use
+    def variables(self) -> List["Variable"]:
         return []
 
     def findall(self, match: Callable[["Expr"], bool]) -> Sequence["Expr"]:
@@ -181,6 +190,7 @@ FALSE = BooleanFalse()
 
 class Not(Expr):
     def __init__(self, expr: Expr) -> None:
+        super().__init__()
         self.expr = expr
 
     def __str__(self) -> str:
@@ -189,8 +199,8 @@ class Not(Expr):
     def __neg__(self) -> Expr:
         return self.expr
 
-    def variables(self, proof: bool = False) -> List["Variable"]:
-        return self.expr.variables(proof)
+    def variables(self) -> List["Variable"]:
+        return self.expr.variables()
 
     @property
     def precedence(self) -> Precedence:
@@ -207,7 +217,8 @@ class Not(Expr):
 
 
 class BinExpr(Expr):
-    def __init__(self, left: Expr, right: Expr) -> None:
+    def __init__(self, left: Expr, right: Expr, location: Location = None) -> None:
+        super().__init__(location)
         self.left = left
         self.right = right
 
@@ -225,8 +236,8 @@ class BinExpr(Expr):
     def precedence(self) -> Precedence:
         raise NotImplementedError
 
-    def variables(self, proof: bool = False) -> List["Variable"]:
-        return list(unique(self.left.variables(proof) + self.right.variables(proof)))
+    def variables(self) -> List["Variable"]:
+        return list(unique(self.left.variables() + self.right.variables()))
 
     def findall(self, match: Callable[["Expr"], bool]) -> Sequence["Expr"]:
         return [
@@ -254,8 +265,9 @@ class BinExpr(Expr):
 
 
 class AssExpr(Expr):
-    def __init__(self, *terms: Expr) -> None:
-        self.terms = list(terms)
+    def __init__(self, *terms: Expr, location: Location = None) -> None:
+        super().__init__(location)
+        self.terms = [t for t in list(terms) if isinstance(t, Expr)]
 
     def __str__(self) -> str:
         if not self.terms:
@@ -306,8 +318,8 @@ class AssExpr(Expr):
     def precedence(self) -> Precedence:
         raise NotImplementedError
 
-    def variables(self, proof: bool = False) -> List["Variable"]:
-        return list(unique([v for t in self.terms for v in t.variables(proof)]))
+    def variables(self) -> List["Variable"]:
+        return list(unique([v for t in self.terms for v in t.variables()]))
 
     def findall(self, match: Callable[["Expr"], bool]) -> Sequence["Expr"]:
         return [
@@ -352,7 +364,7 @@ class AssExpr(Expr):
                 terms.append(Number(total))
         if len(terms) == 1:
             return terms[0]
-        return self.__class__(*terms)
+        return self.__class__(*terms, location=self.location)
 
     @abstractmethod
     def operation(self, left: int, right: int) -> int:
@@ -479,7 +491,8 @@ class OrElse(And):
 
 
 class Number(Expr):
-    def __init__(self, value: int, base: int = 0) -> None:
+    def __init__(self, value: int, base: int = 0, location: Location = None) -> None:
+        super().__init__(location)
         self.value = value
         self.base = base
 
@@ -627,7 +640,7 @@ class Add(AssExpr):
                 terms.insert(0, term)
         if len(terms) == 1:
             return terms[0]
-        return Add(*terms)
+        return Add(*terms, location=self.location)
 
     def neutral_element(self) -> int:
         return 0
@@ -778,7 +791,8 @@ class Mod(BinExpr):
 
 
 class Name(Expr):
-    def __init__(self, negative: bool = False) -> None:
+    def __init__(self, negative: bool = False, location: Location = None) -> None:
+        super().__init__(location)
         self.negative = negative
 
     def __str__(self) -> str:
@@ -816,8 +830,10 @@ class Name(Expr):
 
 
 class Variable(Name):
-    def __init__(self, identifier: StrID, negative: bool = False) -> None:
-        super().__init__(negative)
+    def __init__(
+        self, identifier: StrID, negative: bool = False, location: Location = None
+    ) -> None:
+        super().__init__(negative, location)
         self.identifier = ID(identifier)
 
     @property
@@ -828,7 +844,7 @@ class Variable(Name):
     def representation(self) -> str:
         return str(self.name)
 
-    def variables(self, proof: bool = False) -> List["Variable"]:
+    def variables(self) -> List["Variable"]:
         return [self]
 
     def z3expr(self) -> z3.ArithRef:
@@ -839,12 +855,13 @@ class Variable(Name):
 
 class Attribute(Name):
     def __init__(self, prefix: Union[StrID, Expr], negative: bool = False) -> None:
-        super().__init__(negative)
-
-        if isinstance(prefix, (str, ID)):
+        if isinstance(prefix, ID):
+            prefix = Variable(prefix, location=prefix.location)
+        if isinstance(prefix, str):
             prefix = Variable(prefix)
 
         self.prefix: Expr = prefix
+        super().__init__(negative, prefix.location)
 
     @property
     def representation(self) -> str:
@@ -865,11 +882,9 @@ class Attribute(Name):
         expr = self.__class__(self.prefix.simplified())
         return -expr if self.negative else expr
 
-    def variables(self, proof: bool = False) -> List[Variable]:
-        if proof:
-            if not isinstance(self.prefix, Variable):
-                raise TypeError
-            return [Variable(f"{self.prefix}'{self.__class__.__name__}")]
+    def variables(self) -> List[Variable]:
+        if not isinstance(self.prefix, Variable):
+            raise TypeError
         return self.prefix.variables()
 
     def z3expr(self) -> z3.ArithRef:
@@ -1044,8 +1059,9 @@ UNDEFINED = UndefinedExpr()
 
 
 class Aggregate(Expr):
-    def __init__(self, *elements: Expr) -> None:
-        self.elements = list(elements)
+    def __init__(self, *elements: Expr, location: Location = None) -> None:
+        super().__init__(location)
+        self.elements = [t for t in list(elements) if isinstance(t, Expr)]
 
     def __str__(self) -> str:
         return "(" + ", ".join(map(str, self.elements)) + ")"
@@ -1079,6 +1095,7 @@ class Aggregate(Expr):
 
 class NamedAggregate(Expr):
     def __init__(self, *elements: Tuple[StrID, Expr]) -> None:
+        super().__init__()
         self.elements = [(ID(n), e) for n, e in elements]
 
     def __str__(self) -> str:
@@ -1270,6 +1287,7 @@ class If(Expr):
     def __init__(
         self, condition_expressions: Sequence[Tuple[Expr, Expr]], else_expression: Expr = None
     ) -> None:
+        super().__init__()
         self.condition_expressions = condition_expressions
         self.else_expression = else_expression
 
@@ -1322,13 +1340,13 @@ class If(Expr):
 
         return If(simplified_ce, self.else_expression)
 
-    def variables(self, proof: bool = False) -> List["Variable"]:
+    def variables(self) -> List["Variable"]:
         variables = []
         for ce in self.condition_expressions:
-            variables.extend(ce[0].variables(proof))
-            variables.extend(ce[1].variables(proof))
+            variables.extend(ce[0].variables())
+            variables.extend(ce[1].variables())
         if self.else_expression:
-            variables.extend(self.else_expression.variables(proof))
+            variables.extend(self.else_expression.variables())
         return list(unique(variables))
 
     def z3expr(self) -> z3.ExprRef:
@@ -1352,6 +1370,7 @@ class Case(Expr):
     def __init__(
         self, control_expression: Expr, case_statements: Sequence[Tuple[Expr, Expr]]
     ) -> None:
+        super().__init__()
         self.control_expression = control_expression
         self.case_statements = case_statements
 
@@ -1395,11 +1414,11 @@ class Case(Expr):
             [(c.simplified(), e.simplified()) for c, e in self.case_statements],
         )
 
-    def variables(self, proof: bool = False) -> List["Variable"]:
-        variables = self.control_expression.variables(proof)
+    def variables(self) -> List["Variable"]:
+        variables = self.control_expression.variables()
         for cs in self.case_statements:
-            variables.extend(cs[0].variables(proof))
-            variables.extend(cs[1].variables(proof))
+            variables.extend(cs[0].variables())
+            variables.extend(cs[1].variables())
         return list(unique(variables))
 
     def z3expr(self) -> Union[z3.BoolRef, z3.ExprRef]:
@@ -1421,6 +1440,7 @@ class Case(Expr):
 
 class QuantifiedExpression(Expr):
     def __init__(self, parameter_name: str, iterable: Expr, predicate: Expr) -> None:
+        super().__init__()
         self.parameter_name = parameter_name
         self.iterable = iterable
         self.predicate = predicate
@@ -1451,8 +1471,8 @@ class QuantifiedExpression(Expr):
     def keyword(self) -> str:
         raise NotImplementedError
 
-    def variables(self, proof: bool = False) -> List["Variable"]:
-        return list(unique(self.iterable.variables(proof) + self.predicate.variables(proof)))
+    def variables(self) -> List["Variable"]:
+        return list(unique(self.iterable.variables() + self.predicate.variables()))
 
     def z3expr(self) -> z3.ExprRef:
         raise NotImplementedError
@@ -1486,6 +1506,7 @@ class ForAllIn(QuantifiedExpression):
 
 class ValueRange(Expr):
     def __init__(self, lower: Expr, upper: Expr):
+        super().__init__()
         self.lower = lower
         self.upper = upper
 
