@@ -32,9 +32,11 @@ from rflx.expression import (
     Mod,
     Name,
     NamedAggregate,
+    Not,
     NotEqual,
     Number,
     Or,
+    Relation,
     Selected,
     Size,
     Sub,
@@ -62,9 +64,12 @@ from . import const
 
 
 def substitution(
-    message: Message, embedded: bool = False, public: bool = False
+    message: Message,
+    embedded: bool = False,
+    public: bool = False,
+    target_type: ID = const.TYPES_U64,
 ) -> Callable[[Expr], Expr]:
-    facts = substitution_facts(message, embedded, public)
+    facts = substitution_facts(message, embedded, public, target_type)
 
     def func(expression: Expr) -> Expr:
         def byte_aggregate(aggregate: Aggregate) -> Aggregate:
@@ -77,24 +82,24 @@ def substitution(
             field = None
             aggregate = None
             if isinstance(expression.left, Variable) and isinstance(expression.right, Aggregate):
-                field = expression.left
+                field = Field(expression.left.name)
                 aggregate = byte_aggregate(expression.right)
             elif isinstance(expression.left, Aggregate) and isinstance(expression.right, Variable):
-                field = expression.right
+                field = Field(expression.right.name)
                 aggregate = byte_aggregate(expression.left)
             if field and aggregate:
+                assert field in message.fields
                 if embedded:
                     return Equal(
                         Indexed(
-                            Variable("Buffer"),
+                            Variable("Buffer.all"),
                             ValueRange(
                                 Call(
                                     const.TYPES_BYTE_INDEX,
                                     [
                                         Selected(
                                             Indexed(
-                                                Variable("Cursors"),
-                                                Variable(Field(field.name).affixed_name),
+                                                Variable("Cursors"), Variable(field.affixed_name),
                                             ),
                                             "First",
                                         )
@@ -105,8 +110,7 @@ def substitution(
                                     [
                                         Selected(
                                             Indexed(
-                                                Variable("Cursors"),
-                                                Variable(Field(field.name).affixed_name),
+                                                Variable("Cursors"), Variable(field.affixed_name),
                                             ),
                                             "Last",
                                         )
@@ -116,8 +120,38 @@ def substitution(
                         ),
                         aggregate,
                     )
-                return Call(
-                    "Equal", [Variable("Ctx"), Variable(Field(field.name).affixed_name), aggregate]
+                equal_call = Call(
+                    "Equal", [Variable("Ctx"), Variable(field.affixed_name), aggregate]
+                )
+                return equal_call if isinstance(expression, Equal) else Not(equal_call)
+
+        def field_value(field: Field) -> Expr:
+            if public:
+                return Call(f"Get_{field.name}", [Variable("Ctx")])
+            return Selected(
+                Indexed(
+                    Variable("Ctx.Cursors" if not embedded else "Cursors"),
+                    Variable(field.affixed_name),
+                ),
+                f"Value.{field.name}_Value",
+            )
+
+        if isinstance(expression, Relation):
+            if (
+                isinstance(expression.left, Variable)
+                and Field(expression.left.name) in message.fields
+                and isinstance(expression.right, Number)
+            ):
+                return expression.__class__(
+                    field_value(Field(expression.left.name)), expression.right
+                )
+            if (
+                isinstance(expression.right, Variable)
+                and Field(expression.right.name) in message.fields
+                and isinstance(expression.left, Number)
+            ):
+                return expression.__class__(
+                    expression.left, field_value(Field(expression.right.name))
                 )
 
         return expression
@@ -126,7 +160,10 @@ def substitution(
 
 
 def substitution_facts(
-    message: Message, embedded: bool = False, public: bool = False
+    message: Message,
+    embedded: bool = False,
+    public: bool = False,
+    target_type: ID = const.TYPES_U64,
 ) -> Mapping[Name, Expr]:
     def prefixed(name: str) -> Expr:
         return Variable(f"Ctx.{name}") if not embedded else Variable(name)
@@ -160,11 +197,10 @@ def substitution_facts(
         if isinstance(field_type, Enumeration):
             if public:
                 return Call(
-                    const.TYPES_BIT_LENGTH,
-                    [Call("To_Base", [Call(f"Get_{field.name}", [Variable("Ctx")])])],
+                    target_type, [Call("To_Base", [Call(f"Get_{field.name}", [Variable("Ctx")])])],
                 )
             return Call(
-                const.TYPES_BIT_LENGTH,
+                target_type,
                 [
                     Selected(
                         Indexed(cursors, Variable(field.affixed_name)), f"Value.{field.name}_Value"
@@ -173,9 +209,9 @@ def substitution_facts(
             )
         if isinstance(field_type, Scalar):
             if public:
-                return Call(const.TYPES_BIT_LENGTH, [Call(f"Get_{field.name}", [Variable("Ctx")])])
+                return Call(target_type, [Call(f"Get_{field.name}", [Variable("Ctx")])])
             return Call(
-                const.TYPES_BIT_LENGTH,
+                target_type,
                 [
                     Selected(
                         Indexed(cursors, Variable(field.affixed_name)), f"Value.{field.name}_Value"
@@ -196,15 +232,13 @@ def substitution_facts(
         **{Length(f.name): field_length(f) for f in message.fields},
         **{Variable(f.name): field_value(f, t) for f, t in message.types.items()},
         **{
-            Variable(l): Call(const.TYPES_BIT_LENGTH, [Call("To_Base", [Variable(l)])])
+            Variable(l): Call(target_type, [Call("To_Base", [Variable(l)])])
             for t in message.types.values()
             if isinstance(t, Enumeration)
             for l in t.literals.keys()
         },
         **{
-            Variable(t.package * l): Call(
-                const.TYPES_BIT_LENGTH, [Call("To_Base", [Variable(t.package * l)])]
-            )
+            Variable(t.package * l): Call(target_type, [Call("To_Base", [Variable(t.package * l)])])
             for t in message.types.values()
             if isinstance(t, Enumeration)
             for l in t.literals.keys()
@@ -232,7 +266,9 @@ def message_structure_invariant(
     length = (
         Size(prefix * full_base_type_name(field_type))
         if isinstance(field_type, Scalar)
-        else link.length.substituted(substitution(message, embedded)).simplified()
+        else link.length.substituted(
+            substitution(message, embedded, target_type=const.TYPES_BIT_LENGTH)
+        ).simplified()
     )
     first = (
         prefixed("First")
@@ -333,10 +369,10 @@ def context_predicate(message: Message, composite_fields: Sequence[Field], prefi
                                             ),
                                             Variable(l.source.affixed_name),
                                         ),
-                                        l.condition,
-                                    )
-                                    .substituted(substitution(message, embedded=True))
-                                    .simplified()
+                                        l.condition.substituted(
+                                            substitution(message, embedded=True)
+                                        ),
+                                    ).simplified()
                                     for l in message.incoming(f)
                                 ]
                             ),
@@ -442,7 +478,7 @@ def valid_path_to_next_field_condition(message: Message, field: Field) -> Sequen
         If(
             [
                 (
-                    l.condition,
+                    l.condition.substituted(substitution(message, public=True)),
                     And(
                         Equal(
                             Call(
@@ -456,9 +492,7 @@ def valid_path_to_next_field_condition(message: Message, field: Field) -> Sequen
                     ),
                 )
             ]
-        )
-        .substituted(substitution(message, public=True))
-        .simplified()
+        ).simplified()
         for l in message.outgoing(field)
         if l.target != FINAL
     ]
@@ -592,10 +626,13 @@ def sequence_name(message: Message, field: Field) -> ID:
 
 
 def length_dependent_condition(message: Message) -> bool:
-    return any(
-        [
-            isinstance(v.name, str) and v.name.endswith("'Length")
-            for l in message.structure
-            for v in l.condition.variables(True)
-        ]
+    return (
+        len(
+            [
+                l
+                for link in message.structure
+                for l in link.condition.findall(lambda x: isinstance(x, Length))
+            ]
+        )
+        > 0
     )
