@@ -3,10 +3,10 @@ from rflx.identifier import ID
 
 REFINEMENT_PACKAGE = ID("Contains")
 
+ARITHMETIC_PACKAGE = ID("RFLX_Arithmetic")
 BUILTIN_TYPES_CONVERSIONS_PACKAGE = ID("RFLX_Builtin_Types.Conversions")
 BUILTIN_TYPES_PACKAGE = ID("RFLX_Builtin_Types")
 GENERIC_TYPES_PACKAGE = ID("RFLX_Generic_Types")
-LEMMAS_PACKAGE = ID("RFLX_Lemmas")
 MESSAGE_SEQUENCE_PACKAGE = ID("RFLX_Message_Sequence")
 SCALAR_SEQUENCE_PACKAGE = ID("RFLX_Scalar_Sequence")
 TYPES_PACKAGE = ID("RFLX_Types")
@@ -14,10 +14,10 @@ TYPES_PACKAGE = ID("RFLX_Types")
 LIBRARY_FILES = [
     file_name(str(p)) + ".ads"
     for p in [
+        ARITHMETIC_PACKAGE,
         BUILTIN_TYPES_CONVERSIONS_PACKAGE,
         BUILTIN_TYPES_PACKAGE,
         GENERIC_TYPES_PACKAGE,
-        LEMMAS_PACKAGE,
         MESSAGE_SEQUENCE_PACKAGE,
         SCALAR_SEQUENCE_PACKAGE,
         TYPES_PACKAGE,
@@ -25,8 +25,8 @@ LIBRARY_FILES = [
 ] + [
     file_name(str(p)) + ".adb"
     for p in [
+        ARITHMETIC_PACKAGE,
         GENERIC_TYPES_PACKAGE,
-        LEMMAS_PACKAGE,
         MESSAGE_SEQUENCE_PACKAGE,
         SCALAR_SEQUENCE_PACKAGE,
     ]
@@ -47,3 +47,4 @@ TYPES_FIRST_BIT_INDEX = TYPES * "First_Bit_Index"
 TYPES_LAST_BIT_INDEX = TYPES * "Last_Bit_Index"
 TYPES_OFFSET = TYPES * "Offset"
 TYPES_UNREACHABLE_BIT_LENGTH = TYPES * "Unreachable_Bit_Length"
+TYPES_U64 = TYPES * "U64"
