@@ -40,6 +40,7 @@ from rflx.expression import (
     Indexed,
     Last,
     LessEqual,
+    Mod,
     NamedAggregate,
     Not,
     Number,
@@ -47,6 +48,7 @@ from rflx.expression import (
     Or,
     Range,
     Selected,
+    Size,
     Slice,
     Variable,
 )
@@ -261,13 +263,11 @@ class GeneratorGenerator:
                                         Variable("Ctx"),
                                         Aggregate(
                                             Variable(f.affixed_name),
-                                            Variable("Val")
-                                            if not isinstance(t, Enumeration)
-                                            else Call("To_Base", [Variable("Val")]),
+                                            Call("To_Base", [Variable("Val")]),
                                         ),
                                     ],
                                 ),
-                                Call("Valid", [Variable("Val")])
+                                Call("Valid", [Call("To_Base", [Variable("Val")])])
                                 if not isinstance(t, Enumeration)
                                 else TRUE,
                                 common.sufficient_space_for_field_condition(
@@ -315,10 +315,7 @@ class GeneratorGenerator:
                             ["Field_Value"],
                             "Field_Dependent_Value",
                             Aggregate(
-                                Variable(f.affixed_name),
-                                Variable("Val")
-                                if not isinstance(t, Enumeration)
-                                else Call("To_Base", [Variable("Val")]),
+                                Variable(f.affixed_name), Call("To_Base", [Variable("Val")]),
                             ),
                             True,
                         ),
@@ -380,6 +377,103 @@ class GeneratorGenerator:
             ],
         )
 
+    def create_composite_setter_empty_procedures(self, message: Message) -> UnitPart:
+        def specification(field: Field) -> ProcedureSpecification:
+            return ProcedureSpecification(
+                f"Set_{field.name}_Empty", [InOutParameter(["Ctx"], "Context")]
+            )
+
+        return UnitPart(
+            [
+                *[
+                    SubprogramDeclaration(
+                        specification(f),
+                        [
+                            Precondition(
+                                AndThen(
+                                    *self.setter_preconditions(f),
+                                    *self.unbounded_composite_setter_preconditions(message, f),
+                                    Equal(
+                                        Call(
+                                            "Field_Length",
+                                            [Variable("Ctx"), Variable(f.affixed_name)],
+                                        ),
+                                        Number(0),
+                                    ),
+                                )
+                            ),
+                            Postcondition(And(*self.composite_setter_postconditions(message, f))),
+                        ],
+                    )
+                    for f, t in message.types.items()
+                    if message.is_possibly_empty(f)
+                ],
+            ],
+            [
+                SubprogramBody(
+                    specification(f),
+                    [
+                        ObjectDeclaration(
+                            ["First"],
+                            const.TYPES_BIT_INDEX,
+                            Call("Field_First", [Variable("Ctx"), Variable(f.affixed_name)]),
+                            True,
+                        ),
+                        ObjectDeclaration(
+                            ["Last"],
+                            const.TYPES_BIT_INDEX,
+                            Call("Field_Last", [Variable("Ctx"), Variable(f.affixed_name)]),
+                            True,
+                        ),
+                    ],
+                    [
+                        CallStatement(
+                            "Reset_Dependent_Fields", [Variable("Ctx"), Variable(f.affixed_name)],
+                        ),
+                        Assignment(
+                            "Ctx",
+                            Aggregate(
+                                Variable("Ctx.Buffer_First"),
+                                Variable("Ctx.Buffer_Last"),
+                                Variable("Ctx.First"),
+                                Variable("Last"),
+                                Variable("Ctx.Buffer"),
+                                Variable("Ctx.Cursors"),
+                            ),
+                        ),
+                        Assignment(
+                            Indexed(Variable("Ctx.Cursors"), Variable(f.affixed_name)),
+                            NamedAggregate(
+                                ("State", Variable("S_Valid")),
+                                ("First", Variable("First")),
+                                ("Last", Variable("Last")),
+                                ("Value", NamedAggregate(("Fld", Variable(f.affixed_name)))),
+                                (
+                                    "Predecessor",
+                                    Selected(
+                                        Indexed(Variable("Ctx.Cursors"), Variable(f.affixed_name)),
+                                        "Predecessor",
+                                    ),
+                                ),
+                            ),
+                        ),
+                        Assignment(
+                            Indexed(
+                                Variable("Ctx.Cursors"),
+                                Call("Successor", [Variable("Ctx"), Variable(f.affixed_name)]),
+                            ),
+                            NamedAggregate(
+                                ("State", Variable("S_Invalid")),
+                                ("Predecessor", Variable(f.affixed_name)),
+                            ),
+                        ),
+                    ],
+                )
+                for f, t in message.types.items()
+                if message.is_possibly_empty(f)
+            ],
+        )
+
     def create_composite_setter_procedures(self, message: Message) -> UnitPart:
         def specification(field: Field) -> ProcedureSpecification:
             return ProcedureSpecification(f"Set_{field.name}", [InOutParameter(["Ctx"], "Context")])
@@ -396,7 +490,12 @@ class GeneratorGenerator:
                     ProcedureSpecification(
                         f"Process_{field.name}", [OutParameter([field.name], const.TYPES_BYTES)],
                     )
-                )
+                ),
+                FormalSubprogramDeclaration(
+                    FunctionSpecification(
+                        "Valid_Length", "Boolean", [Parameter(["Length"], const.TYPES_LENGTH)],
+                    )
+                ),
             ]
 
         return UnitPart(
@@ -408,6 +507,23 @@ class GeneratorGenerator:
                             AndThen(
                                 *self.setter_preconditions(f),
                                 *self.unbounded_composite_setter_preconditions(message, f),
+                                Call(
+                                    "Valid_Length",
+                                    [
+                                        Call(
+                                            const.TYPES_LENGTH,
+                                            [
+                                                Div(
+                                                    Call(
+                                                        "Field_Length",
+                                                        [Variable("Ctx"), Variable(f.affixed_name)],
+                                                    ),
+                                                    Size(const.TYPES_BYTE),
+                                                ),
+                                            ],
+                                        ),
+                                    ],
+                                ),
                             )
                         ),
                         Postcondition(And(*self.composite_setter_postconditions(message, f),)),
@@ -425,6 +541,15 @@ class GeneratorGenerator:
                             AndThen(
                                 *self.setter_preconditions(f),
                                 *self.bounded_composite_setter_preconditions(message, f),
+                                Call(
+                                    "Valid_Length",
+                                    [
+                                        Call(
+                                            const.TYPES_LENGTH,
+                                            [Div(Variable("Length"), Size(const.TYPES_BYTE))],
+                                        )
+                                    ],
+                                ),
                             )
                         ),
                         Postcondition(And(*self.composite_setter_postconditions(message, f),)),
@@ -647,6 +772,20 @@ class GeneratorGenerator:
                 ),
             ),
             common.sufficient_space_for_field_condition(Variable(field.affixed_name)),
+            Equal(
+                Mod(
+                    Call("Field_First", [Variable("Ctx"), Variable(field.affixed_name)]),
+                    Size(const.TYPES_BYTE),
+                ),
+                Number(1),
+            ),
+            Equal(
+                Mod(
+                    Call("Field_Length", [Variable("Ctx"), Variable(field.affixed_name)]),
+                    Size(const.TYPES_BYTE),
+                ),
+                Number(0),
+            ),
         ]
 
     @staticmethod
@@ -688,6 +827,14 @@ class GeneratorGenerator:
                     if Last("Message") in l.length
                 ]
             ),
+            Equal(
+                Mod(
+                    Call("Field_First", [Variable("Ctx"), Variable(field.affixed_name)]),
+                    Size(const.TYPES_BYTE),
+                ),
+                Number(1),
+            ),
+            Equal(Mod(Variable("Length"), Size(const.TYPES_BYTE)), Number(0),),
         ]
 
 
