@@ -8,6 +8,7 @@ from pyparsing import (
     Literal,
     Optional,
     ParseFatalException,
+    ParserElement,
     ParseResults,
     QuotedString,
     Regex,
@@ -22,10 +23,12 @@ from pyparsing import (
     alphas,
     delimitedList,
     infixNotation,
+    locatedExpr,
     nums,
     opAssoc,
 )
 
+from rflx.error import Location, RecordFluxError, Severity, Subsystem, fail, parser_location
 from rflx.expression import (
     TRUE,
     UNDEFINED,
@@ -33,6 +36,7 @@ from rflx.expression import (
     Aggregate,
     And,
     Attribute,
+    BooleanTrue,
     Div,
     Equal,
     Expr,
@@ -53,18 +57,10 @@ from rflx.expression import (
     Variable,
 )
 from rflx.identifier import ID
-from rflx.model import (
-    Array,
-    Enumeration,
-    ModelError,
-    ModularInteger,
-    RangeInteger,
-    Type,
-    is_builtin_type,
-    qualified_type_name,
-)
+from rflx.model import Enumeration, ModularInteger, RangeInteger, Type, qualified_type_name
 
 from .ast import (
+    ArraySpec,
     Component,
     ContextSpec,
     DerivationSpec,
@@ -76,6 +72,8 @@ from .ast import (
     Then,
 )
 
+ParserElement.enablePackrat()
+
 
 def comma() -> Token:
     return Suppress(Literal(",")).setName('","')
@@ -87,7 +85,7 @@ def semicolon() -> Token:
 
 def unqualified_identifier() -> Token:
     return (
-        (WordStart(alphas) + Word(alphanums + "_") + WordEnd(alphanums + "_"))
+        locatedExpr(WordStart(alphas) + Word(alphanums + "_") + WordEnd(alphanums + "_"))
         .setParseAction(verify_identifier)
         .setName("Identifier")
     )
@@ -96,7 +94,12 @@ def unqualified_identifier() -> Token:
 def qualified_identifier() -> Token:
     return (
         Optional(unqualified_identifier() + Literal(".")) + unqualified_identifier()
-    ).setParseAction(lambda t: "".join(t.asList()))
+    ).setParseAction(
+        lambda t: ID(
+            "".join(map(str, t.asList())),
+            Location(start=t[0].location.start, end=t[-1].location.end),
+        )
+    )
 
 
 def attribute_reference() -> Token:
@@ -120,10 +123,11 @@ def numeric_literal() -> Token:
     based_literal = numeral + Literal("#") - based_numeral - Literal("#")
     based_literal.setParseAction(lambda t: (int(t[2], int(t[0])), int(t[0])))
 
-    return (
-        (based_literal | decimal_literal)
-        .setParseAction(lambda t: Number(t[0][0], t[0][1]))
-        .setName("Number")
+    num_literal = based_literal | decimal_literal
+    num_literal.setName("Number")
+
+    return locatedExpr(num_literal).setParseAction(
+        lambda s, l, t: Number(t[0][1][0], t[0][1][1], parser_location(t[0][0], t[0][2], s))
     )
 
 
@@ -148,10 +152,10 @@ def mathematical_expression() -> Token:
     highest_precedence_operator = Literal("**")
 
     array_aggregate = (
-        Suppress(Literal("("))
+        Literal("(").setParseAction(lambda s, l, t: l)
         + numeric_literal()
         + (comma() - numeric_literal()) * (0,)
-        + Suppress(Literal(")"))
+        + Literal(")").setParseAction(lambda s, l, t: l)
     )
     array_aggregate.setParseAction(parse_array_aggregate)
 
@@ -240,7 +244,7 @@ def enumeration_type_definition() -> Token:
 
 
 def array_type_definition() -> Token:
-    return (Keyword("array of") + unqualified_identifier()).setName("Array")
+    return (Keyword("array of") + qualified_identifier()).setName("Array")
 
 
 def message_type_definition() -> Token:
@@ -251,7 +255,7 @@ def message_type_definition() -> Token:
     component_aspects = Keyword("with") - delimitedList(first_aspect | length_aspect)
     component_aspects.setParseAction(parse_aspects)
 
-    then = (
+    then = locatedExpr(
         Keyword("then")
         - (Keyword("null") | unqualified_identifier())
         - Group(Optional(component_aspects))
@@ -274,7 +278,7 @@ def message_type_definition() -> Token:
     )
     component_item.setName("Component")
     null_component_item = Keyword("null") - then - semicolon()
-    null_component_item.setParseAction(lambda t: Component("null", "null", [t[1]]))
+    null_component_item.setParseAction(lambda t: Component(None, None, [t[1]]))
     null_component_item.setName("NullComponent")
     component_list = Group(
         Optional(null_component_item) - component_item - ZeroOrMore(component_item)
@@ -296,14 +300,18 @@ def type_declaration() -> Token:
     )
 
     return (
-        Keyword("type") - unqualified_identifier() - Keyword("is") - type_definition - semicolon()
+        Keyword("type").setParseAction(lambda s, l, t: l)
+        - unqualified_identifier()
+        - Keyword("is")
+        - type_definition
+        - semicolon().setParseAction(lambda s, l, t: l)
     ).setParseAction(parse_type)
 
 
 def type_refinement() -> Token:
     return (
         (
-            Suppress(Keyword("for"))
+            Suppress(Keyword("for")).setParseAction(lambda s, l, t: l)
             - qualified_identifier()
             - Suppress(Keyword("use"))
             - Suppress(Literal("("))
@@ -312,7 +320,7 @@ def type_refinement() -> Token:
             - qualified_identifier()
             - Suppress(Literal(")"))
             - Optional(value_constraint())("constraint")
-            - semicolon()
+            - semicolon().setParseAction(lambda s, l, t: l)
         )
         .setParseAction(parse_refinement)
         .setName("Refinement")
@@ -330,7 +338,7 @@ def package_declaration() -> Token:
         - Keyword("end")
         - unqualified_identifier()
         - semicolon()
-    ).setParseAction(parse_package_declaration)
+    ).setParseAction(lambda t: PackageSpec(t[1], t[3].asList(), t[5]))
 
 
 def context_clause() -> Token:
@@ -351,7 +359,7 @@ def fatalexceptions(parse_function: Callable) -> Callable:
     def wrapper(string: str, location: int, tokens: ParseResults) -> object:
         try:
             return parse_function(string, location, tokens)
-        except ParseFatalException as e:
+        except (ParseFatalException, RecordFluxError) as e:
             raise e
         except Exception as e:
             raise ParseFatalException(string, location, f"implementation error ({e})")
@@ -361,23 +369,12 @@ def fatalexceptions(parse_function: Callable) -> Callable:
 
 @fatalexceptions
 def parse_array_aggregate(string: str, location: int, tokens: ParseResults) -> Expr:
-    check_aggregate_elements(tokens, string, location)
-    return Aggregate(*tokens)
+    return Aggregate(*tokens[1:-1], location=parser_location(tokens[0], tokens[-1], string))
 
 
 @fatalexceptions
 def parse_string(string: str, location: int, tokens: ParseResults) -> Expr:
-    elements = [Number(ord(c)) for c in tokens[0]]
-    check_aggregate_elements(elements, string, location)
-    return Aggregate(*elements)
-
-
-def check_aggregate_elements(elements: List[Number], string: str, location: int) -> None:
-    for element in elements:
-        if not Number(0) <= element <= Number(255):
-            raise ParseFatalException(
-                string, location, f'Number "{element}" is out of range 0 .. 255'
-            )
+    return Aggregate(*[Number(ord(c)) for c in tokens[0]])
 
 
 @fatalexceptions
@@ -387,25 +384,28 @@ def parse_concatenation(string: str, location: int, tokens: ParseResults) -> Exp
 
 @fatalexceptions
 def parse_term(string: str, location: int, tokens: ParseResults) -> Expr:
-    if isinstance(tokens[0], str):
-        return Variable(tokens[0])
+    if isinstance(tokens[0], ID):
+        return Variable(tokens[0], negative=False, location=tokens[0].location)
     return tokens[0]
 
 
 @fatalexceptions
 def parse_relation(string: str, location: int, tokens: ParseResults) -> Relation:
+    def locn() -> Location:
+        return Location(tokens[0].location.start, tokens[0].location.source, tokens[2].location.end)
+
     if tokens[1] == "<":
-        return Less(tokens[0], tokens[2])
+        return Less(tokens[0], tokens[2], locn())
     if tokens[1] == "<=":
-        return LessEqual(tokens[0], tokens[2])
+        return LessEqual(tokens[0], tokens[2], locn())
     if tokens[1] == "=":
-        return Equal(tokens[0], tokens[2])
+        return Equal(tokens[0], tokens[2], locn())
     if tokens[1] == ">=":
-        return GreaterEqual(tokens[0], tokens[2])
+        return GreaterEqual(tokens[0], tokens[2], locn())
     if tokens[1] == ">":
-        return Greater(tokens[0], tokens[2])
+        return Greater(tokens[0], tokens[2], locn())
     if tokens[1] == "/=":
-        return NotEqual(tokens[0], tokens[2])
+        return NotEqual(tokens[0], tokens[2], locn())
     raise ParseFatalException(string, location, "unexpected relation operator")
 
 
@@ -417,10 +417,13 @@ def parse_logical_expression(string: str, location: int, tokens: ParseResults) -
         operator = result.pop(0)
         right = result.pop(0)
         expression: Expr
+        assert left.location
+        assert right.location
+        locn = Location(left.location.start, left.location.source, right.location.end)
         if operator == "and":
-            expression = And(left, right)
+            expression = And(left, right, location=locn)
         elif operator == "or":
-            expression = Or(left, right)
+            expression = Or(left, right, location=locn)
         else:
             raise ParseFatalException(string, location, "unexpected logical operator")
         result.insert(0, expression)
@@ -434,17 +437,23 @@ def parse_mathematical_expression(string: str, location: int, tokens: ParseResul
         left = result.pop(0)
         operator = result.pop(0)
         right = result.pop(0)
+        assert left.location, f'expression "{left}" without location'
+        assert right.location, f'expression "{right}" without location'
+        assert left.location.source == right.location.source, "expression with different source"
+        locn = Location(left.location.start, left.location.source, left.location.end)
         expression: Expr
         if operator == "+":
             expression = Add(left, right)
+            expression.location = locn
         elif operator == "-":
-            expression = Sub(left, right)
+            expression = Sub(left, right, locn)
         elif operator == "*":
             expression = Mul(left, right)
+            expression.location = locn
         elif operator == "/":
-            expression = Div(left, right)
+            expression = Div(left, right, locn)
         elif operator == "**":
-            expression = Pow(left, right)
+            expression = Pow(left, right, locn)
         else:
             raise ParseFatalException(string, location, "unexpected mathematical operator")
         result.insert(0, expression)
@@ -453,16 +462,20 @@ def parse_mathematical_expression(string: str, location: int, tokens: ParseResul
 
 @fatalexceptions
 def parse_then(string: str, location: int, tokens: ParseResults) -> Then:
+    tokens = tokens[0]
+    start = tokens.pop(0)
+    locn = parser_location(start, tokens[-1], string)
     return Then(
-        tokens[1],
+        tokens[1] if tokens[1] != "null" else None,
         tokens[2][0]["first"] if tokens[2] and "first" in tokens[2][0] else UNDEFINED,
         tokens[2][0]["length"] if tokens[2] and "length" in tokens[2][0] else UNDEFINED,
-        tokens[3][0] if tokens[3] else TRUE,
+        tokens[3][0] if tokens[3] else BooleanTrue(location=locn),
+        locn,
     )
 
 
 @fatalexceptions
-def verify_identifier(string: str, location: int, tokens: ParseResults) -> str:
+def verify_identifier(string: str, location: int, tokens: ParseResults) -> ID:
     reserved_words = [
         "abort",
         "abs",
@@ -540,11 +553,19 @@ def verify_identifier(string: str, location: int, tokens: ParseResults) -> str:
         "initial",
         "final",
     ]
+    data = tokens[0].asDict()
+    tokens = data["value"]
+    locn = parser_location(data["locn_start"], data["locn_end"], string)
+
     if tokens[0].lower() in reserved_words:
-        raise ParseFatalException(
-            string, location, f'reserved word "{tokens[0]}" used as identifier'
+        fail(
+            f'reserved word "{tokens[0]}" used as identifier',
+            Subsystem.PARSER,
+            Severity.ERROR,
+            locn,
         )
-    return tokens[0]
+
+    return ID(tokens[0], locn)
 
 
 @fatalexceptions
@@ -571,60 +592,48 @@ def parse_aspects(string: str, location: int, tokens: ParseResults) -> Dict[str,
 
 @fatalexceptions
 def parse_type(string: str, location: int, tokens: ParseResults) -> Type:
-    try:
-        package = ID("__PACKAGE__")
-        name = tokens[1]
+    package = ID("__PACKAGE__")
+    name = tokens[1]
 
-        if is_builtin_type(name):
-            raise ParseFatalException(
-                string, location, f'illegal redefinition of built-in type "{name}"'
-            )
+    locn = parser_location(tokens[0], tokens[-1], string)
 
-        identifier = package * name
+    identifier = package * name
 
-        if tokens[3] == "mod":
-            return ModularInteger(identifier, *tokens[4:6])
-        if tokens[3] == "range":
-            tokens[6] = tokens[6]["size"]
-            return RangeInteger(identifier, *tokens[4:7])
-        if tokens[3] == "message":
-            return MessageSpec(identifier, tokens[4])
-        if tokens[3] == "null message":
-            return MessageSpec(identifier, [])
-        if tokens[3] == "(":
-            elements = dict(tokens[4:-2])
-            aspects = tokens[-1]
-            if len(elements) < len(tokens[4:-2]):
-                raise ModelError(f'"{name}" contains duplicate elements')
-            if "always_valid" not in aspects:
-                aspects["always_valid"] = False
-            return Enumeration(identifier, elements, aspects["size"], aspects["always_valid"])
-        if tokens[3] == "new":
-            return DerivationSpec(identifier, tokens[4])
-        if tokens[3] == "array of":
-            return Array(identifier, ReferenceSpec(qualified_type_name(tokens[4], package)))
+    if tokens[3] == "mod":
+        return ModularInteger(identifier, tokens[4], locn)
+    if tokens[3] == "range":
+        tokens[6] = tokens[6]["size"]
+        return RangeInteger(identifier, tokens[4], tokens[5], tokens[6], locn)
+    if tokens[3] == "message":
+        return MessageSpec(identifier, tokens[4], locn)
+    if tokens[3] == "null message":
+        return MessageSpec(identifier, [], locn)
+    if tokens[3] == "(":
+        elements = tokens[4:-3]
+        aspects = tokens[-2]
+        if "always_valid" not in aspects:
+            aspects["always_valid"] = False
+        enumeration = Enumeration(
+            identifier, elements, aspects["size"], aspects["always_valid"], locn
+        )
+        return enumeration
+    if tokens[3] == "new":
+        return DerivationSpec(identifier, tokens[4], locn)
+    if tokens[3] == "array of":
+        return ArraySpec(
+            identifier,
+            ReferenceSpec(qualified_type_name(tokens[4], package), tokens[4].location),
+            locn,
+        )
 
-    except ModelError as e:
-        raise ParseFatalException(string, location, e)
     raise ParseFatalException(string, location, "unexpected type")
 
 
 @fatalexceptions
 def parse_refinement(string: str, location: int, tokens: ParseResults) -> RefinementSpec:
-    if "constraint" not in tokens:
-        tokens.append(TRUE)
-    return RefinementSpec(tokens[0], tokens[1], tokens[2], tokens[3])
-
-
-@fatalexceptions
-def parse_package_declaration(string: str, location: int, tokens: ParseResults) -> PackageSpec:
-    if tokens[1].startswith("RFLX"):
-        raise ParseFatalException(
-            string, location, f'illegal prefix "RFLX" in package identifier "{tokens[1]}"'
-        )
-    if tokens[1] != tokens[5]:
-        raise ParseFatalException(string, location, "inconsistent package identifiers")
-    return PackageSpec(tokens[1], tokens[3].asList())
+    constraint = tokens[4] if "constraint" in tokens else TRUE
+    locn = parser_location(tokens[0], tokens[-1], string)
+    return RefinementSpec(tokens[1], tokens[2], tokens[3], constraint, locn)
 
 
 def unit() -> Token:
